# Duo, A crazy eights game.
# Copyright (C) 2003-2008 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Game scheduler and arbiter.
"""

import logging

from duo import protocol
from duo import game


class GameAlreadyRunningError(Exception):
    """Error raised on multiple game start."""
    pass

class InvalidMoveError(Exception):
    """Error raised on invalid move."""
    pass


# TODO: needs refactoring.
class PlayerInfo(object):
    def __init__(self, name):
        self.name = name
        self.hand = []
        self.ready = False
        self.drawn = False


class PlayersRing(list):
    """The players list considering their order and their hands.

    Every entry of this list is a tuple containing first the player name
    and then this player hand and its status.
    """

    # Game directions
    CW = +1
    CCW = -1

    def __init__(self):
        list.__init__(self)
        self._direction = PlayersRing.CW

        # Current player index is negative until the first player is elected.
        self._current = -1
        self._dealer = -1

    def add(self, player):
        """Add a new player."""
        if player in [pinfo.name for pinfo in self]:
            raise protocol.PlayerAlreadyAuthenticatedError
        self.append(PlayerInfo(player))

    def remove(self, player):
        """Remove a player."""
        for pinfo in self:
            if pinfo.name == player:
                self.remove(pinfo)
                return

    def get_list(self):
        """Return the player names."""
        return [pinfo.name for pinfo in self]

    def get_list_except(self, player):
        """Retrieve the list of player, minus the one given."""
        return [pinfo.name for pinfo in self if pinfo.name != player]

    def update_player_status(self, player, ready=None, drawn=None):
        """Update the status of the given player."""
        for pinfo in self:
            if pinfo.name == player:
                if not ready is None:
                    pinfo.ready = ready
                if not drawn is None:
                    pinfo.drawn = drawn
                break

    def reset(self):
        """Reset the status of all players."""
        self._dealer = (self._dealer + 1) % len(self)
        self._current = -1
        self._direction = PlayersRing.CW
        for pinfo in self:
            pinfo.ready = False
            pinfo.drawn = False
            pinfo.hand = []

    def get_player_hand(self, player):
        """Retrieve the hand of the given player."""
        entries = [pinfo.hand for pinfo in self if pinfo.name == player]
        if entries:
            return entries[0]
        else:
            return None

    def has_player_drawn(self, player):
        """Check if the given player has already drawn a card."""
        entries = [pinfo.drawn for pinfo in self if pinfo.name == player]
        if entries:
            return entries[0]
        else:
            return None

    def current(self):
        """Return the current player."""
        return self[self._current].name if self._current >= 0 else None

    def next(self):
        """Return the next player to play."""
        return self[self._next_current()].name

    def change_direction(self):
        """Change the direction of progression in the players ring."""
        self._direction = - self._direction

    def pass_turn(self):
        """Remember to pass for next player turn computation."""
        self._direction = 2 * self._direction

    def get_deal_order(self):
        """Return the ordered indexes of the players for the next game.

        Order is determined by the dealer position. The first player
        of the list if the first on the left of the dealer. The rest
        of the list is build with the rest of the players, clockwise.
        """
        deal_order = []
        for i in xrange(len(self)):
            deal_order.append((i + 1 + self._dealer) % len(self))
        return deal_order

    def compute_next(self):
        """Compute a new TurnMessage with the next player to play."""
        # Reset the drawing status of the current player.
        self[self._current].drawn = False

        self._current = self._next_current()
        self._direction = self._direction / abs(self._direction)
        return protocol.TurnMessage(self.current())

    def is_ready(self):
        """Return True if all players are ready, false otherwise."""
        for pinfo in self:
            if pinfo.ready == False:
                return False
        return True

    def _next_current(self):
        """Return the next current pointer value."""
        if self._current >= 0:
            return (self._current + self._direction) % len(self)
        else:
            # Compute initial player...
            return (self._dealer + self._direction) % len(self)


DEFAULT_RULESET = ['TurnRule', 'ReverseRule', 'PassRule',
                   'DrawTwoRule', 'DrawFourRule',
                   'ColorRule', 'TypeRule']


class Rule(object):
    """The referee rules base class.

    When considering a move, the referee tries to find a matching rule
    in its rule set. Rules are tested using their check method in
    order of appearance in the rule set. On success, the rule apply
    method is invoked.

    Rules checking shall return True on success, False otherwise. A
    rule can also raise an exception if not checked: the move is
    invalid and no further check is done.
    """
    def __init__(self, referee):
        self._referee = referee

    def _get_referee(self):
        """Retrieve the referee."""
        return self._referee

    referee = property(_get_referee, None, None, 'The referee')

    def check(self, player, card):
        """Check this rule."""
        return False

    def apply(self, message_bundle):
        """Apply this rule."""
        pass


class ColorRule(Rule):
    """Rule allowing to play a card with the same color."""
    def __init__(self, referee):
        Rule.__init__(self, referee)

    def check(self, player, card):
        """Check this rule.

        This rule should match only if the given card color matches
        the heap top card color, or if it is a jocker (a not colored
        card). It MUST fail if a kitty is present or was just drawn by
        the player.
        """
        # Hack for letting the first card be returned
        if player == None:
            return True
        kitty = self.referee.get_kitty() 
        if len(kitty) or kitty.taken:
            return False
        return (
            card.color == self.referee.get_current_color()
            or card.color == 'N')


class TurnRule(Rule):
    """Rule enforcing player turns."""
    def __init__(self, referee):
        Rule.__init__(self, referee)

    def check(self, player, card):
        """Check this rule.

        This rule should match only if the given player is the player
        to play.
        """
        if player:
            if player != self.referee.get_players().current():
                raise InvalidMoveError()
        return False


class TypeRule(Rule):
    """Rule allowing to play a card with the same type."""
    def __init__(self, referee):
        Rule.__init__(self, referee)

    def check(self, player, card):
        """Check this rule.

        This rule should match only if the given card type matches the
        heap top card type. It MUST fail if a kitty is present or was
        just drawn by the player.
        """
        # Hack for letting the first card be returned
        if player == None:
            return True
        kitty = self.referee.get_kitty() 
        if len(kitty) or kitty.taken:
            return False
        return card.type == self.referee.get_heap().peek().type


class ReverseRule(Rule):
    """Rule recognizing the reverse card."""
    def __init__(self, referee):
        Rule.__init__(self, referee)

    def check(self, player, card):
        """Check this rule.

        This rule should match only if the given card is a Reverse
        card, and type or color matches with the heap top. It MUST
        fail if a kitty is present or was just drawn by the player.
        """
        kitty = self.referee.get_kitty() 
        if len(kitty) or kitty.taken:
            return False
        top = self.referee.get_heap().peek()
        return card.type == 'R' and (
            top == None or
            card.type == top.type or
            card.color == self.referee.get_current_color())

    def apply(self, message_bundle):
        """Apply this rule."""
        # in case of a two-player game, 'R' card behavior is equivalent to 'P'
        if len(self.referee.get_players()) == 2:
            self.referee.get_players().pass_turn()
        else:
            self.referee.get_players().change_direction()
        message_bundle.append(protocol.ReverseMessage())


class PassRule(Rule):
    """Rule allowing to pass."""
    def __init__(self, referee):
        Rule.__init__(self, referee)

    def check(self, player, card):
        """Check this rule.

        This rule should match only if the given card is a Pass card,
        and type or color matches with the heap top. It MUST fail if a
        kitty is present or was just drawn by the player.
        """
        kitty = self.referee.get_kitty() 
        if len(kitty) or kitty.taken:
            return False
        top = self.referee.get_heap().peek()
        return card.type == 'P' and (
            top == None or
            card.type == top.type or
            card.color == self.referee.get_current_color())

    def apply(self, message_bundle):
        """Apply this rule."""
        self.referee.get_players().pass_turn()
        message_bundle.append(protocol.PassMessage())


class DrawTwoRule(Rule):
    """Rule enforcing the Draw Two cards card."""
    def __init__(self, referee):
        Rule.__init__(self, referee)

    def check(self, player, card):
        """Check this rule.

        This rule should match only if the given card is a Draw Two
        Cards card and type or color matches with the heap top. It
        MUST fail if a kitty was initiated with a Draw Four card or
        the kitty was just picked up.
        """
        kitty = self.referee.get_kitty()
        if kitty.type == 'X' or kitty.taken:
            return False
        
        top = self.referee.get_heap().peek()
        return card.type == 'D' and (
            top == None or
            card.type == top.type or
            card.color == self.referee.get_current_color())

    def apply(self, message_bundle):
        """Apply this rule."""
        kitty = self.referee.get_kitty() 
        kitty.type = 'D'
        for times in range(2):
            kitty.append(self.referee.get_deck().pop())


class DrawFourRule(Rule):
    """Rule enforcing the Draw Far Cards card."""
    def __init__(self, referee):
        Rule.__init__(self, referee)

    def check(self, player, card):
        """Check this rule.

        This rule should match only if the given card is a Draw For
        Cards card. It MUST fail if a kitty was initiated with a Draw
        Two card or the kitty was just picked up.
        """
        kitty = self.referee.get_kitty()
        if kitty.type == 'D' or kitty.taken:
            return False
        return card.type == 'X'

    def apply(self, message_bundle):
        """Apply this rule."""
        kitty = self.referee.get_kitty() 
        kitty.type = 'X'
        for times in range(4):
            kitty.append(self.referee.get_deck().pop())


#REFEREE_OPTIONS = {
#    'game': ('stupid-default-name', 'The name to use when opening the game'),
#    'cards': (7, 'The number of cards to deal'),
#    'ruleset': ('DEFAULT_RULESET', 'The ruleset to use')
#}

class Referee(protocol.ProtocolAdapter):
    """The game referee.
    """
    WAITING = 'waiting'
    DEALING = 'dealing'
    PLAYING = 'playing'

    # TODO: nbCards and Ruleset should become referee options.
    def __init__(self, transport_layer, transport_options, referee_options):
        protocol.ProtocolAdapter.__init__(self, 'referee')
        self.connect(transport_layer, transport_options)

        self.register(protocol.StartMessage, self.on_start_cb, None)
        self.register(protocol.ReadyMessage, self.on_ready_cb, None)
        self.register(protocol.PlayMessage, self.on_play_cb, None)
        self.register(protocol.DrawMessage, self.on_draw_cb, None)
        self.register(protocol.PassMessage, self.on_pass_cb, None)

        self._players = PlayersRing()
        self._options = referee_options
        self._ruleset = [
            eval(rule + '(self)') for rule in eval(referee_options['ruleset'])]

        self._current_color = None
        self._game_status = Referee.WAITING

        self._deck = None
        self._heap = None
        self._kitty = None
        logging.getLogger('duo.referee').info('Referee ready.')

    def leave(self):
        """Leave and finish the game"""
        # Escape twice to leave the game, then the transport layer.
        self._endpoint.escape()
        self._endpoint.escape()

    def send_except(self, message, player):
        """Send the given message to everyone except the given player."""
        for player_ in self._players.get_list_except(player):
            self.send_to(message, player_)

    def on_endpoint_join(self, player, emitter):
        """Callback invoked by the transport layer on new connection."""
        if self._game_status != Referee.WAITING:
            raise GameAlreadyRunningError
        self._players.add(player)
        logging.getLogger('duo.referee').info('%s joined the game', player)
        self.broadcast(protocol.AuthMessage(self._players.get_list()))

    def on_endpoint_escape(self, player, emitter):
        """Callback invoked by the transport layer on disconnection."""
        # TODO: if game is running, should probably be interrupted
        self._players.remove(player)
        logging.getLogger('duo.referee').info('%s left the game', player)

    def on_start_cb(self, key, player, args):
        """Callback invoked on reception of a Start message."""
        if key == self._options['master_key']:
            self._players.reset()
            self._deck = game.Deck()
            self._heap = game.Deck()
            self._kitty = game.Kitty()
            self._deal()
        else:
            logging.getLogger('duo.referee').warning(
                'received start with wrong key')

    def on_ready_cb(self, message, player, args):
        """Callback invoked on reception of a Ready message."""
        if self._game_status != Referee.DEALING:
            raise GameAlreadyRunningError

        self._players.update_player_status(player, ready=True)
        if self._players.is_ready():
            self._start_game()

    def on_play_cb(self, message, player, args):
        """Callback invoked on reception of a Play message."""
        if self._game_status != Referee.PLAYING:
            raise InvalidMoveError('All players must be ready first.')
        card, color = message
        self._play(player, card, color)

    def on_draw_cb(self, message, player, args):
        """Callback invoked on reception of a Draw message."""
        # TODO: Should proceed to some checks like the ruleset on card play.
        if self._game_status != Referee.PLAYING:
            raise InvalidMoveError('All players must be ready first.')
        try:
            cards = []

            if self._players.has_player_drawn(player):
                raise InvalidMoveError('Already drawn once')

            if len(self._kitty):
                cards.extend(self._kitty.pick_up())
            else:
                cards.append(self._deck.pop())
            self._players.get_player_hand(player).extend(cards)
            self._players.update_player_status(player, drawn=True)

            self.send_to(protocol.HandMessage(
                self._players.get_player_hand(player)), player)
            self.broadcast(protocol.DealMessage(len(cards), player))
        except Exception, ex:
            self.send_to(protocol.ErrorMessage(ex), player)

    def on_pass_cb(self, message, player, args):
        """Callback invoked on reception of a Pass message."""
        # TODO: Should proceed to some checks like the ruleset on card play.
        if self._kitty.taken:
            self._kitty = game.Kitty()
        self.broadcast(self._players.compute_next())

    def get_players(self):
        """Return the players list."""
        return self._players

    def get_kitty(self):
        """Return the current pick."""
        return self._kitty

    def get_heap(self):
        """Return the heap."""
        return self._heap

    def get_deck(self):
        """Return the deck."""
        return self._deck

    def get_current_color(self):
        """Return the currently asked color."""
        return self._current_color

    def _deal(self):
        """Proceed to the first cards distribution."""
        logging.getLogger('duo.referee').info('Dealing cards')
        self._game_status = Referee.DEALING
        self._deck.generate()
        self._deck.shuffle()

        # Deal the cards
        hands = [[] for player in range(len(self._players))]
        for times in range(self._options['cards']):
            for player in range(len(hands)):
                hands[player].append(self._deck.pop())

        for i in self._players.get_deal_order():
            self._players[i].hand.extend(hands[i])
            self.send_to(protocol.HandMessage(hands[i]), self._players[i].name)

    def _start_game(self):
        """Start the game.

        Game start is invoked when all the players are ready. It is
        achieved by returning the card of the deck onto the heap, and
        applying its effects if any.
        """
        logging.getLogger('duo.referee').info('Starting game')
        self._game_status = Referee.PLAYING

        # Set the heap top card...
        drawn_card = self._deck.pop()
        logging.getLogger('duo.referee').debug('drawing %s', drawn_card)
        # Should the following be dropped ?
        while drawn_card.type in [None, 'X', 'W']:
            self._heap.append(drawn_card)
            drawn_card = self._deck.pop()
            logging.getLogger('duo.referee').debug('drawing %s', drawn_card)

        # ... And apply its effects
        self._play(None, drawn_card)

    def _play(self, player, card, color=None):
        """Check or reject the last move."""
        # TODO: Add the Big Lock
        try:
            if not player is None and \
                   not card in self._players.get_player_hand(player):
                raise InvalidMoveError('Invalid card. Are you a cheater ?')

            for rule in self._ruleset:
                logging.getLogger('duo.referee').debug(
                    'Referee.move checking %s', rule.__class__.__name__)
                if rule.check(player, card):
                    self._do_play(player, card, color, rule)
                    return

            raise InvalidMoveError('No rule applied')

        except Exception, ex:
            self.send_to(protocol.ErrorMessage(ex), player)

    def _do_play(self, player, card, color, rule):
        """Apply the effects of the last move."""
        logging.getLogger('duo.referee').debug(
            'Referee._do_play applying %s', rule.__class__.__name__)
        message_bundle = protocol.MessageBundle()
        rule.apply(message_bundle)
        self._heap.append(card)
        if card.color != 'N':
            self._current_color = self._heap.peek().color
        else:
            self._current_color = color

        if player != None:
            self._players.get_player_hand(player).remove(card)
            message_bundle.append(
                protocol.HeapMessage(self._heap.peek(),
                                     self._current_color, player))

            # Check victory.
            if self._players.get_player_hand(player) != []:
                message_bundle.append(self._players.compute_next())
            self.broadcast(message_bundle)

            # For Tk style, the hand message must come after the Heap
            # information.
            self.send_to(
                protocol.HandMessage(
                self._players.get_player_hand(player)), player)

            # The WinnerMessage must be the very last.
            if self._players.get_player_hand(player) == []:
                self.broadcast(protocol.WinnerMessage(player))

        else:
            # Player is None: it's the first move
            message_bundle.append(
                protocol.HeapMessage(
                self._heap.peek(), self._current_color, player))
            message_bundle.append(self._players.compute_next())
            self.broadcast(message_bundle)
